""" Image Resource
    ---
    ```
    from realpy import RsImage
    ```
"""
import os
from typing import Sequence, overload, AnyStr

import pygame.image as PyImage
import pygame.rect as PyRect

__all__ = ["RsImage"]


class RsImage(object):
    """`RsImage(image_path)`
        ---
        Raw image of a sprite that contains single or multiple images.
    """

    @overload
    def __init__(self, filepath: AnyStr):
        ...

    @overload
    def __init__(self, filepath: Sequence[AnyStr]):
        ...

    def __init__(self, filepath):
        self.number: int = -1
        self.raw_data = []
        self.boundbox = PyRect.Rect(0, 0, 0, 0)

        if isinstance(filepath, str):
            filepath = (filepath,)

        self.number = len(filepath)
        if self.number == 0:
            raise FileNotFoundError

        FileLoc: str
        for FileLoc in filepath:
            self.raw_data.append(PyImage.load(FileLoc).convert_alpha())

        FileLoc = filepath[0]
        self.filename = os.path.basename(FileLoc)
        self.boundbox = self.raw_data[0].get_rect()



import asyncio
import sys

import pygame
import pygame.constants as PyConstants
import pygame.display as PyDisplay
import pygame.fastevent as PyEvent
import pygame.mixer as PyAudio
from pygame.time import Clock

__all__ = ["rs_init", "rs_startup", "rs_quit"]

mouse_igniter = []
key_igniter = []
controller_igniter = []
mouse_proceed = []
key_proceed = []
controller_proceed = []
other_proceed = []


def proceed_ing(where, pack):
    from realpy.core import INPUT_STATES, debug_get

    while True:
        try:
            Upk = where.pop()
            if pack.get(Upk) == INPUT_STATES.RELEASED:
                pack[Upk] = INPUT_STATES.NONE
                continue
            pack[Upk] = INPUT_STATES.ING
            if debug_get():
                print("Start at later:", Upk)
        except IndexError:
            break
    where.clear()


def proceed_done(where, pack):
    from realpy.core import INPUT_STATES, debug_get

    while True:
        try:
            Upk = where.pop()
            pack[Upk] = INPUT_STATES.NONE
            if debug_get():
                print("Cleaned at later:", Upk)
        except IndexError:
            break
    where.clear()


async def update_hand():
    from realpy.core.constants import INPUT_STATES
    from realpy.core.preset import RsPreset

    RsPreset.event_others.clear()

    # Await
    Temp = PyEvent.get()
    Len = len(Temp)

    proceed_ing(mouse_igniter, RsPreset.event_mouse)
    proceed_ing(key_igniter, RsPreset.event_key)
    proceed_ing(controller_igniter, RsPreset.event_controller)
    proceed_done(mouse_proceed, RsPreset.event_mouse)
    proceed_done(key_proceed, RsPreset.event_key)
    proceed_done(controller_proceed, RsPreset.event_controller)
    # await asyncio.gather(proceed_ing(mouse_igniter, RsPreset.event_mouse), ...)

    if 0 < Len:
        for event in Temp:
            if event.type == PyConstants.QUIT:
                raise RsPreset.RsInteruptError

            elif event.type == PyConstants.KEYDOWN:
                Seed: int = event.key
                Place = RsPreset.event_key.get(Seed)

                if Place:  # Key can be able to be None
                    if Place == INPUT_STATES.NONE:  # Normal
                        RsPreset.event_key[Seed] = INPUT_STATES.PRESSED
                    elif Place == INPUT_STATES.RELEASED:  # Should not happen
                        RsPreset.event_key[Seed] = INPUT_STATES.NONE
                    else:  # It may not be runned
                        RsPreset.event_key[Seed] = INPUT_STATES.ING
                else:  # Add a new key
                    key_igniter.append(Seed)
                    RsPreset.event_key[Seed] = INPUT_STATES.PRESSED
                    if RsPreset.debug_get():
                        print("New Key:", Seed)

            elif event.type == PyConstants.KEYUP:
                Seed = event.key
                Place = RsPreset.event_key.get(Seed)

                if Place:
                    if Place == INPUT_STATES.RELEASED:  # Should not happen
                        RsPreset.event_key[Seed] = INPUT_STATES.NONE
                    elif Place != INPUT_STATES.NONE:
                        key_proceed.append(Seed)  # Clear later
                        RsPreset.event_key[Seed] = INPUT_STATES.RELEASED
                else:
                    RsPreset.event_key[Seed] = INPUT_STATES.NONE

            elif event.type == PyConstants.MOUSEBUTTONDOWN:
                Seed: int = event.button
                Place = RsPreset.event_mouse.get(Seed)
                RsPreset.mouse_x, RsPreset.mouse_y = event.pos

                if Place:
                    if Place == INPUT_STATES.NONE:
                        RsPreset.event_mouse[Seed] = INPUT_STATES.PRESSED
                    elif Place == INPUT_STATES.RELEASED:
                        RsPreset.event_mouse[Seed] = INPUT_STATES.NONE
                    else:
                        RsPreset.event_mouse[Seed] = INPUT_STATES.ING
                else:
                    mouse_igniter.append(Seed)
                    RsPreset.event_mouse[Seed] = INPUT_STATES.PRESSED
                    if RsPreset.debug_get():
                        print("New Click:", Seed)

            elif event.type == PyConstants.MOUSEBUTTONUP:
                Seed = event.button
                Place = RsPreset.event_mouse.get(Seed)
                RsPreset.mouse_x, RsPreset.mouse_y = event.pos

                if Place:
                    if Place == INPUT_STATES.RELEASED:
                        RsPreset.event_mouse[Seed] = INPUT_STATES.NONE
                    elif Place != INPUT_STATES.NONE:
                        mouse_proceed.append(Seed)  # Clear later
                        RsPreset.event_mouse[Seed] = INPUT_STATES.RELEASED
                else:
                    RsPreset.event_mouse[Seed] = INPUT_STATES.NONE

            elif event.type == PyConstants.MOUSEMOTION:
                RsPreset.mouse_x, RsPreset.mouse_y = event.pos
            else:
                RsPreset.event_others.append(event)


async def update_flow(room, time: float):
    room.onUpdate(time)
    room.onUpdateLater(time)


async def update_draw(room, time: float):
    from realpy.core.preset import RsPreset

    RsPreset.application_surface.fill("black")
    room.onDraw(time)
    PyDisplay.update()


async def update_all(room, time: float):
    await asyncio.gather(update_flow(room, time), update_draw(room, time))


def rs_init(title: str, view_port_width: int, view_port_height: int, *, audio_channels=12, audio_buffer=1024):
    from realpy.core.preset import RsPreset

    PyAudio.pre_init(channels=audio_channels, buffer=audio_buffer)
    pygame.init()
    PyDisplay.init()
    PyEvent.init()
    PyAudio.init()

    PyDisplay.set_caption(title)
    PyDisplay.set_allow_screensaver(False)

    RsPreset.dimension = (view_port_width, view_port_height)
    RsPreset.application_surface = PyDisplay.set_mode(RsPreset.dimension)


def rs_startup():
    from realpy.core.preset import RsPreset
    from realpy.core.scene import RsScene

    # Startup
    Rooms = RsPreset.room_order
    StartRoom = None
    try:
        StartRoom = Rooms[0]
        RsPreset.room = StartRoom
    except IndexError:
        raise RuntimeError("No scene found.")

    if not StartRoom:
        raise RuntimeError("Invalid scene.")

    RoomCurrent: RsScene = RsPreset.room
    AbsoluteTimer = Clock()
    TimeOccured: float = 0

    # Load rooms
    RsPreset.room.onAwake()
    while True:
        TimeOccured = 0 if RsPreset.room.paused else AbsoluteTimer.get_time() * 0.001  # Millisecond

        try:
            asyncio.run(update_hand())
            asyncio.run(update_all(RoomCurrent, TimeOccured))
        except RsPreset.RsInteruptError:
            rs_quit()

        AbsoluteTimer.tick(RsPreset.game_speed)
        RoomCurrent = RsPreset.room


def rs_quit():
    print("Program is ended.")
    PyAudio.quit()
    pygame.quit()
    sys.exit(0)



from .constants import INPUT_STATES
from .preset import RsPreset


def keyboard_check(key: int) -> bool:
    Place = RsPreset.event_key.get(key)
    if Place and Place == INPUT_STATES.ING or Place == INPUT_STATES.PRESSED:
        return True
    else:
        return False


def keyboard_check_pressed(key: int) -> bool:
    Place = RsPreset.event_key.get(key)
    if Place and Place == INPUT_STATES.PRESSED:
        return True
    else:
        return False


def keyboard_check_released(key: int) -> bool:
    Place = RsPreset.event_key.get(key)
    if Place and Place == INPUT_STATES.RELEASED:
        return True
    else:
        return False


def mouse_check(button: int) -> bool:
    Place = RsPreset.event_mouse.get(button)
    if Place and Place == INPUT_STATES.ING or Place == INPUT_STATES.PRESSED:
        return True
    else:
        return False


def mouse_check_pressed(button: int) -> bool:
    Place = RsPreset.event_mouse.get(button)
    if Place and Place == INPUT_STATES.PRESSED:
        return True
    else:
        return False


def mouse_check_released(button: int) -> bool:
    Place = RsPreset.event_mouse.get(button)
    if Place and Place == INPUT_STATES.RELEASED:
        return True
    else:
        return False



""" Scene
    ---
    ```
    from realpy import RsScene
    ```
"""
from typing import Optional, Any, List, Dict


class RsScene(object):
    """`RsScene(name)`
        ---
        Large portion of game pipeline.
    """

    def __init__(self, name: str = ""):
        self.name: str = name
        self.layer_stack = []
        self.trees: Dict[str, Any] = {}
        self.paused: bool = False
        self.EveryInstancesPot = []
        self.SpecificInstancesPot: Dict[int, List] = {}
        self.before: Optional[RsScene] = None
        self.next: Optional[RsScene] = None

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Realpy Scene {self.name} ({len(self.layer_stack)})"

    def add_layer_direct(self, layer):
        self.layer_stack.append(layer)
        self.trees[layer.name] = layer
        return layer

    def layer_find(self, caption: str):
        return self.trees.get(caption)

    def pause(self) -> None:
        self.paused = True

    def resume(self) -> None:
        self.paused = False

    def onAwake(self) -> None:
        for Layer in self.layer_stack:
            Layer.onAwake()

    def onDestroy(self) -> None:
        for Layer in self.layer_stack:
            Layer.onDestroy()

    def onUpdate(self, time: float) -> None:
        for Layer in self.layer_stack:
            Layer.onUpdate(time)

    def onUpdateLater(self, time: float) -> None:
        for Layer in self.layer_stack:
            Layer.onUpdateLater(time)

    def onDraw(self, time: float) -> None:
        for Layer in self.layer_stack:
            Layer.onDraw(time)



""" Layer
    ---
    ```
    from realpy import RsLayer
    ```
"""


class RsLayer(object):
    """`RsLayer(name)`
        ---
        Belongs to a scene and executes game instances.
    """

    def __init__(self, scene, name: str = ""):
        from typing import List, Optional

        self.scene = scene
        self.name: str = name
        self.storage: List = []
        self.storage_awake: List = []
        self.storage_destroy: List = []
        self.storage_update: List = []
        self.storage_updatelater: List = []
        self.storage_draw: List = []
        self.__atomic_storage: Optional[List] = None

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Realpy Layer {self.name} ({len(self.storage)})"

    def add(self, instance):
        instance.attach(self.storage)

        if instance.onAwake:
            instance.attach(self.storage_awake)

        if instance.onDestroy:
            instance.attach(self.storage_destroy)

        if instance.onUpdate:
            instance.attach(self.storage_update)

        if instance.onUpdateLater:
            instance.attach(self.storage_updatelater)

        if instance.onDraw:
            instance.attach(self.storage_draw)

        return instance

    def onAwake(self) -> None:
        for Instance in self.storage_awake:
            Instance.onAwake()

    def onDestroy(self) -> None:
        for Instance in self.storage_destroy:
            Instance.onDestroy()

    def onUpdate(self, time: float) -> None:
        from copy import copy

        self.__atomic_storage = copy(self.storage_update)
        for Instance in self.__atomic_storage:
            Instance.onUpdate(time)

    def onUpdateLater(self, time: float) -> None:
        from copy import copy

        self.__atomic_storage = copy(self.storage_updatelater)
        for Instance in self.__atomic_storage:
            Instance.onUpdateLater(time)

    def onDraw(self, time: float) -> None:
        from copy import copy

        self.__atomic_storage = copy(self.storage_draw)
        for Instance in self.__atomic_storage:
            Instance.onDraw(time)



from .header import (
    PySound, PyChannel, PyMusic, RsSound,
    audio_play, audio_stop, audio_play_single, audio_pause, audio_resume,
    audio_is_playing, audio_set_channel_count, audio_get_channel_count
)



from copy import copy
from typing import Any, List, Optional, Tuple, Type


class RsInstance(object):
    """`RsInstance(Prefab, Layer, x, y)`
        ---
    """

    def __init__(self, original: Type[Any], layer, x: float = 0, y: float = 0):
        from pygame.surface import Surface
        from realpy.core.layer import RsLayer
        from realpy.core.sprite import RsSprite

        self.enabled: bool = True
        self.layer: Optional[RsLayer] = layer
        self.department: List[List] = []
        self.visible: bool = True
        self.original = original
        self.x: float = x
        self.y: float = y
        self.__use_collision: bool = original.use_collision
        self.__sprite_index: Optional[RsSprite] = original.sprite_index
        self.__image: Optional[Surface] = None
        self.image_index: float = 0
        self.__image_angle: float = 0
        self.image_scale: float = 1
        self.image_alpha: float = 1
        self.__speed: float = 0
        self.__direction: float = 0
        self.__hspeed: float = 0
        self.__vspeed: float = 0
        self.friction: float = 0
        self.boundbox: Optional[List[Tuple[int, int]]]
        self.bound_vertexes: Optional[List[Tuple[int, int]]]
        if self.__sprite_index:
            self.boundbox = copy(self.__sprite_index.boundbox)
            self.bound_vertexes = copy(self.__sprite_index.boundbox)
        else:
            self.boundbox = None
            self.bound_vertexes = None

        MethodA = self.original.onAwake
        if MethodA:
            self.onAwake = lambda: MethodA(self)
        else:
            self.onAwake = None
        MethodD = self.original.onDestroy
        if MethodD:
            self.onDestroy = lambda: MethodD(self)
        else:
            self.onDestroy = None
        Method = self.original.onUpdate
        if Method:
            self.onUpdate = lambda time: Method(self, time)
        else:
            self.onUpdate = None

    def __del__(self):
        for Group in self.department:
            Group.remove(self)

    def attach(self, ogroup: List):
        ogroup.append(self)
        self.department.append(ogroup)

    def acceleration(self, velocity, direction):
        from realpy.core.arithmetic import lengthdir_x, lengthdir_y, point_distance, point_direction

        Vx = lengthdir_x(velocity, direction)
        Vy = lengthdir_y(velocity, direction)
        self.__hspeed += Vx
        self.__vspeed += Vy
        self.__speed = point_distance(0, 0, self.__hspeed, self.__vspeed)
        self.__direction = point_direction(0, 0, self.__hspeed, self.__vspeed)

    def draw_self(self) -> bool:
        from realpy import RsPreset

        Where = RsPreset.application_surface
        if Where and self.sprite_index:
            self.__image = self.sprite_index.draw(Where, self.image_index, self.x, self.y,
                                                  self.image_scale, self.image_angle, self.image_alpha).convert_alpha()

            if RsPreset.debug_get() and self.can_collide:
                from pygame import draw
                draw.line(Where, "red", self.bound_vertexes[0], self.bound_vertexes[1])
                draw.line(Where, "red", self.bound_vertexes[1], self.bound_vertexes[3])
                draw.line(Where, "red", self.bound_vertexes[3], self.bound_vertexes[2])
                draw.line(Where, "red", self.bound_vertexes[2], self.bound_vertexes[0])
                draw.circle(Where, "red", (self.x, self.y), 8)
            return True
        else:
            return False

    @property
    def can_collide(self) -> bool:
        return self.__use_collision

    @property
    def sprite_index(self):
        return self.__sprite_index

    @property
    def current_image(self):
        return self.__image

    @property
    def speed(self) -> float:
        return self.__speed

    @property
    def direction(self) -> float:
        return self.__direction

    @property
    def hspeed(self) -> float:
        return self.__hspeed

    @property
    def vspeed(self) -> float:
        return self.__vspeed

    @property
    def image_angle(self) -> float:
        return self.__image_angle

    @speed.setter
    def speed(self, value: float):
        from realpy.core.arithmetic import lengthdir_x, lengthdir_y

        self.__speed = value
        self.__hspeed = lengthdir_x(value, self.__direction)
        self.__vspeed = lengthdir_y(value, self.__direction)

    @direction.setter
    def direction(self, value: float):
        from realpy.core.arithmetic import lengthdir_x, lengthdir_y

        self.__direction = value
        if self.__speed != 0:
            self.__hspeed = lengthdir_x(self.__speed, self.__direction)
            self.__vspeed = lengthdir_y(self.__speed, self.__direction)

    @hspeed.setter
    def hspeed(self, value: float):
        from realpy.core.arithmetic import point_distance, point_direction

        self.__hspeed = value
        self.__speed = point_distance(0, 0, self.__hspeed, self.__vspeed)
        self.__direction = point_direction(0, 0, self.__hspeed, self.__vspeed)

    @vspeed.setter
    def vspeed(self, value: float):
        from realpy.core.arithmetic import point_distance, point_direction

        self.__vspeed = value
        self.__speed = point_distance(0, 0, self.__hspeed, self.__vspeed)
        self.__direction = point_direction(0, 0, self.__hspeed, self.__vspeed)

    @sprite_index.setter
    def sprite_index(self, index):
        # Update the original boundbox
        if not index:  # Free the memory
            self.__sprite_index = None
            del self.__image
            del self.boundbox
            del self.bound_vertexes
        elif not self.__sprite_index:  # Create boundbox
            self.__sprite_index = index
            if self.can_collide:  # Use only at it can collide with other
                self.boundbox = copy(index.boundbox)
                self.bound_vertexes = copy(index.boundbox)
        else:  # Don't update the currrent boundbox
            self.__sprite_index = index

    @image_angle.setter
    def image_angle(self, value: float):
        # Update the actual boundbox
        if self.__image_angle != value:
            self.__image_angle = value

    def onUpdateLater(self, time: float) -> None:
        """`onUpdateLater(time)`
            ---
            Do not override it.
        """
        from realpy import RsPreset
        from realpy.core.arithmetic import lengthdir_x, lengthdir_y

        Method = self.original.onUpdateLater
        if Method:
            Method(self, time)

        if self.speed != 0:
            Hspeed: float
            Vspeed: float
            if self.friction != 0:
                Fx = lengthdir_x(self.friction, self.direction)
                Fy = lengthdir_y(self.friction, self.direction)
                if 0 < self.speed:
                    Hspeed = self.__hspeed * time - 0.5 * time * time * Fx
                    Vspeed = self.__vspeed * time - 0.5 * time * time * Fy
                    self.speed = max(0, self.speed - self.friction * time)
                else:
                    Hspeed = self.__hspeed * time + 0.5 * time * time * Fx
                    Vspeed = self.__vspeed * time + 0.5 * time * time * Fy
                    self.speed = min(0, self.speed + self.friction * time)
            else:
                Hspeed = self.__hspeed * time
                Vspeed = self.__vspeed * time

            if Hspeed != 0:
                self.x += Hspeed

            if Vspeed != 0:
                self.y += Vspeed

        if self.can_collide and self.__sprite_index and RsPreset.debug_get():
            self._set_vertex_boundary(self.__image_angle)

    def onDraw(self, time: float) -> None:
        """`onDraw(time)`
            ---
            Do not override it.
        """
        if self.visible:
            Method = self.original.onDraw
            if Method:
                Method(self, time)
            else:
                self.draw_self()

    def __str__(self) -> str:
        return f"Realpy Instance of {self.original}"

    def __repr__(self) -> str:
        return f"Instance {id(self)} at '{self.layer}'"

    def _set_vertex_boundary(self, angle: float) -> None:
        from realpy.core.arithmetic import lengthdir_x, lengthdir_y

        Cos = lengthdir_x(self.image_scale, angle)
        Sin = lengthdir_y(self.image_scale, angle)

        for i in range(0, 4):
            TempPoint = self.boundbox[i]
            TempA = TempPoint[0]
            TempB = TempPoint[1]
            TempX = round(self.x + TempA * Cos - TempB * Sin)
            TempY = round(self.y + TempA * Sin + TempB * Cos)
            # print("Point(", i, ") - ", TempPoint, " -> (", TempX, ", ", TempY, ")")

            self.bound_vertexes[i] = (TempX, TempY)



""" Arithmetic
    ---
    ```
    from realpy.core import arithmetic
    ```
"""
import math
import random

__all__ = [
    "sqr", "sign", "degtorad", "radtodeg", "irandom", "irandom_range", "bezier4", "choose",
    "distribute", "probability_test", "lengthdir_x", "lengthdir_y", "line_interact",
    "point_distance", "point_direction"
]


def sqr(v: float) -> float:
    return v * v


def sign(x: float) -> int:
    ret = 0
    if x > 0:
        ret = 1
    elif x < 0:
        ret = - 1
    return ret


def degtorad(degree: float) -> float:
    return math.radians(degree)


def radtodeg(radian: float) -> float:
    return math.degrees(radian)


def bezier4(t: float, x1, x2, x3, x4) -> float:
    factor = 1 - t

    return factor * (factor * (factor * x1 + t * x2)
                     + t * (factor * x2 + t * x3)) + t * (factor * (factor * x2 + t * x3) + t * (factor * x3 + t * x4))


def lengthdir_x(length: float, direction: float) -> float:
    return math.cos(degtorad(direction)) * length


def lengthdir_y(length: float, direction: float):
    return -math.sin(degtorad(direction)) * length


def point_distance(x1: float, y1: float, x2: float, y2: float) -> float:
    return math.dist([x1, y1], [x2, y2])


def point_direction(x1: float, y1: float, x2: float, y2: float) -> float:
    return radtodeg(math.atan2(y2 - y1, x1 - x2))


def line_interact(Sx1: float, Sy1: float, Sx2: float, Sy2: float, Dx1: float, Dy1: float, Dx2: float, Dy2: float,
                  Seg: bool) -> float:
    Sx = Sx2 - Sx1
    Sy = Sy2 - Sy1
    Dx = Dx2 - Dx1
    Dy = Dy2 - Dy1
    wx = Sx1 - Dx1
    wy = Sy1 - Dy1

    ua = 0
    ud = Dy * Sx - Dx * Sy

    if ud != 0:
        ua = (Dx * wy - Dy * wx) / ud
        if Seg:
            if ua < 0 or ua > 1:
                return 0

            ub = (Sx * wy - Sy * wx) / ud
            if ub < 0 or ub > 1:
                return 0
    return ua


def irandom(n) -> int:
    return random.randint(0, int(n))


def irandom_range(n1, n2) -> int:
    return random.randint(int(n1), int(n2))


def distribute(x1: float, x2: float, ratio: float) -> float:
    if irandom(100) <= ratio * 100:
        return x1
    else:
        return x2


def probability_test(max_value) -> bool:
    return bool(irandom(max_value - 1) == 0)


def choose(*args):
    length = len(args)
    if length <= 0:
        raise RuntimeError("choose 함수에 값이 제대로 전달되지 않았습니다!" + __name__)

    pick = None
    try:
        pick = args[irandom(length - 1)]
    except ValueError:
        pass
    return pick



from typing import Optional, List, Dict

from pygame.mixer import Sound as PyAudio
from pygame.surface import Surface as PySurface

from ..scene import RsScene

__all__ = [
    "game_speed", "mouse_x", "mouse_y",
    "dimension", "application_surface",
    "room", "room_last", "room_order", "room_all",
    "event_mouse", "event_key", "event_controller", "event_others", "RsInteruptError",
    "debug_set", "debug_get"
]

# General
game_speed: int = 30
mouse_x: int = 0
mouse_y: int = 0

_realpy_debug: bool = False


def debug_set(flag: bool):
    global _realpy_debug
    _realpy_debug = flag


def debug_get() -> bool:
    global _realpy_debug
    return _realpy_debug


# Display
dimension = (640, 480)

# Application surface
application_surface: Optional[PySurface] = None

# Rooms
room: Optional[RsScene] = None
room_last: Optional[RsScene] = None
room_order: List[RsScene] = []
room_all: Dict[str, RsScene] = {}

# Events
event_mouse: Dict[int, int] = {}
event_key: Dict[int, int] = {}
event_controller: Dict[int, int] = {}
event_others = []

# Others
AudioPot: Dict[str, PyAudio] = {}


class RsInteruptError(Exception):
    pass



from typing import List, Optional, Type, overload

import pygame.mask as PyMask

from .behavior import RsPrefab, RsActor, RsInstance
from .layer import RsLayer
from .preset import RsPreset


def collide_anyone(instance: RsInstance, prefab: Type[RsPrefab]) -> Optional[RsInstance]:
    TempHash = hash(prefab)
    Pot: List
    try:
        Pot = RsPreset.room.SpecificInstancesPot[TempHash]
    except KeyError:
        return None

    if not instance.enabled or not instance.visible or not instance.can_collide or not instance.current_image:
        return None

    Checker = PyMask.from_surface(instance.current_image)

    Other: RsInstance
    for Other in Pot:
        if Other == instance:
            continue
        if not Other.enabled or not Other.visible or not Other.can_collide or not Other.current_image:
            continue

        OtherChecker = PyMask.from_surface(Other.current_image)
        Check = Checker.overlap(OtherChecker, (int(Other.x - instance.x), int(Other.y - instance.y)))
        if Check:
            return Other

    return None


def collide_all(instance: RsInstance, prefab: Type[RsPrefab]) -> Optional[List[RsInstance]]:
    TempHash = hash(prefab)
    Pot: List
    try:
        Pot = RsPreset.room.SpecificInstancesPot[TempHash]
    except KeyError:
        return None

    if len(Pot) == 0 or not instance.enabled or not instance.visible or not instance.can_collide or not \
            instance.current_image:
        return None

    Result: List[RsInstance] = []
    Checker = PyMask.from_surface(instance.current_image)

    Other: RsInstance
    for Other in Pot:
        if Other == instance:
            continue
        if not Other.enabled or not Other.visible or not Other.can_collide or not Other.current_image:
            continue

        OtherChecker = PyMask.from_surface(Other.current_image)
        Check = Checker.overlap(OtherChecker, (int(Other.x - instance.x), int(Other.y - instance.y)))
        if Check:
            Result.append(Other)

    return Result


@overload
def actor_create(actor_type: Type[RsActor], layer_id: Optional[str]) -> RsActor: ...


@overload
def actor_create(actor_type: Type[RsActor], layer_id: Optional[RsLayer]) -> RsActor: ...


def actor_create(actor_type: Type[RsActor], layer_id) -> RsActor:
    """`actor_create(Scene, Layer)`
        ---
        Creates an simple behavior instance.
    """

    Instance: RsActor
    if layer_id:
        Place: Optional[RsLayer]
        if isinstance(layer_id, RsLayer):
            Place = layer_id
        else:
            Place = RsPreset.room.layer_find(layer_id)

        if Place:
            Instance = actor_type(Place)
            Method = Instance.onAwake
            if Method:
                Method()
            Place.add(Instance)
        else:
            raise RuntimeError("The specific layer not exists.")
    else:
        Instance = actor_type(None)
        Method = Instance.onAwake
        if Method:
            Method()

    return Instance


@overload
def instance_create(prefab: Type[RsPrefab], layer_id: Optional[str], x: float = ..., y: float = ...) -> RsInstance: ...


@overload
def instance_create(prefab: Type[RsPrefab], layer_id: Optional[RsLayer], x: float = ...,
                    y: float = ...) -> RsInstance: ...


def instance_create(prefab: Type[RsPrefab], layer_id, x: float = 0, y: float = 0) -> RsInstance:
    """`instance_create(Scene, Layer, x=0, y=0)`
        ---
        Creates an instance of game object.
    """

    Instance: RsInstance
    if layer_id:
        Place: Optional[RsLayer]
        if isinstance(layer_id, RsLayer):
            Place = layer_id
        else:
            Place = RsPreset.room.layer_find(layer_id)

        if Place:
            Instance = prefab.trait_instance(prefab, Place, x, y)
            Place.add(Instance)
        else:
            raise RuntimeError(f"The specific layer '{layer_id}' not exists.")
    else:
        Instance = prefab.trait_instance(prefab, None, x, y)

    Method = Instance.onAwake
    if Method:
        Method()

    _instance_register(prefab, Instance)

    return Instance


def instance_number(prefab: Type[RsPrefab]) -> int:
    TempHash = hash(prefab)
    Pot: List
    try:
        Pot = RsPreset.room.SpecificInstancesPot[TempHash]
    except KeyError:
        return 0
    return len(Pot)


def instance_find(prefab: Type[RsPrefab], index: int) -> Optional[RsInstance]:
    TempHash = hash(prefab)
    Pot: List
    try:
        Pot = RsPreset.room.SpecificInstancesPot[TempHash]
    except KeyError:
        return None
    return Pot[index]


def _instance_register(prefab: Type[RsPrefab], instance: RsInstance):
    if issubclass(prefab, RsPrefab) and prefab is not RsPrefab:
        instance.attach(RsPreset.room.EveryInstancesPot)
        _instance_register_recursive(prefab, instance)


def _instance_register_recursive(prefab: Type[RsPrefab], instance: RsInstance):
    TempHash = hash(prefab)
    Where: List
    try:
        Where = RsPreset.room.SpecificInstancesPot[TempHash]
    except KeyError:
        RsPreset.room.SpecificInstancesPot[TempHash] = []
        Where = RsPreset.room.SpecificInstancesPot[TempHash]

    instance.attach(Where)

    Recursive_condition = (issubclass(prefab, RsPrefab) and prefab is not RsPrefab)
    if RsPreset.debug_get():
        print(Recursive_condition, "from", prefab)
        print(f"An instance is appended to {Where} of {prefab}.")
        if Recursive_condition:
            print("Goto ", prefab.__base__)

    if Recursive_condition:
        _instance_register(prefab.__base__, instance)


@overload
def instance_destroy(target: RsActor): ...


@overload
def instance_destroy(target: RsInstance): ...


def instance_destroy(target):
    Method = target.onDestroy
    if Method:
        Method()
    del target



""" Prefab
    ---
    ```
    from realpy import RsPrefab, RsActor, RsInstance
    ```
"""
from .actor import RsActor
from .instance import RsInstance

__all__ = ["RsPrefab", "RsActor", "RsInstance"]


class RsPrefab(object):
    """`RsPrefab`
        ---
        Predefined behavior object.

        Do not instantiate it.

        ### List of flow methods
        - `onAwake()`
        - `onDestroy()`
        - `onUpdate(delta_time)`
        - `onUpdateLater(delta_time)`
        - `onDraw(delta_time)`

        ### Specification
        ```
        @staticmethod
        def onDraw(itself, time: float) -> None:
            pass
        ```
    """

    name: str = ""
    sprite_index = None
    use_collision: bool = True
    trait_instance = RsInstance

    onAwake = None
    onDestroy = None
    onUpdate = None
    onUpdateLater = None
    onDraw = None

    @staticmethod
    def onTest(itself, time: float) -> None:
        """`onTest(instance, time)`
            ---
            This will run on its instance. You may override it.
        """
        pass



from typing import Optional, overload

from pygame import mixer as PyAudio
from pygame.mixer import music as PyMusic, Channel as PyChannel, Sound as PySound

from .sfx import RsSound

audio_set_channel_count = PyAudio.set_num_channels
audio_get_channel_count = PyAudio.get_num_channels


def audio_play(sound: RsSound, loop=False):
    return sound.play(-1) if loop else sound.play()


def audio_stop(sound: Optional[RsSound] = None):
    if sound:
        sound.stop()
    else:
        PyAudio.stop()


def audio_play_single(sound: RsSound, loop=False):
    sound.stop()
    audio_play(sound, loop)


def audio_pause(sound: Optional[RsSound] = None):
    if sound:
        if sound.is_playing():
            for Place in sound.fields:
                if Place.get_sound() is sound:
                    Place.pause()
    else:
        PyAudio.pause()


def audio_resume(sound: Optional[RsSound] = None):
    if sound:
        for Place in sound.fields:
            if Place.get_sound() is sound:
                Place.unpause()
    else:
        PyAudio.unpause()


@overload
def audio_is_playing(sound: Optional[PySound]) -> bool: ...


@overload
def audio_is_playing(sound: Optional[PyChannel]) -> bool: ...


def audio_is_playing(sound=None) -> bool:
    if sound is None:
        return PyAudio.get_busy()
    elif isinstance(sound, PySound):
        return 0 < sound.get_num_channels()
    elif isinstance(sound, PyChannel):
        return sound.get_busy()
    else:
        raise TypeError(f"{sound} is not sound element.")


@overload
def audio_fadeout(time: int, sound: Optional[PySound]): ...


@overload
def audio_fadeout(time: int, sound: Optional[PyChannel]): ...


def audio_fadeout(time: int, sound=None):
    if sound is None:
        PyAudio.fadeout(time)
    elif isinstance(sound, PySound):
        sound.fadeout(time)
    elif isinstance(sound, PyChannel):
        sound.fadeout(time)
    else:
        raise TypeError(f"{sound} is not sound element.")



# TODO: #38 Collision field: New tile based collider
from typing import List, overload


class RsCollsionField(object):
    def __init__(self, x: int, y: int, width: int, height: int, grid_width: int, grid_height: int):
        self.enabled = True
        self.X = x
        self.Y = y
        self.WIDTH = width
        self.HEIGHT = height
        self.GRID_W: int = grid_width
        self.GRID_H: int = grid_height

        count_w = int(width / grid_width)
        count_h = int(height / grid_height)
        self.cells: List[int] = [0] * count_h * count_w

    @overload
    def set(self, x: int, y: int, value: int):
        self.cells[x + y * self.WIDTH] = value

    @overload
    def set(self, x: int, y: int, container: List[int]):
        for nth, value in enumerate(container):
            self.cells[x + y * self.WIDTH + nth] = value

    def get(self, x: int, y: int) -> int:
        return self.cells[x + y * self.WIDTH]



# 10 px == 1 metre
phy_mess: float = 10.0 / 1
phy_velocity: float = (((1000.0 / 60.0) / 60.0) * phy_mess)

# Mouse buttons
MB_LEFT = 1
MB_MIDDLE = 2
MB_RIGHT = 3
MB_WHEEL = MB_MIDDLE
MB_WHEEL_UP = 4
MB_WHEEL_DOWN = 5

# Virtual keys
VK_LEFT = 1073741904
VK_RIGHT = 1073741903
VK_UP = 1073741906
VK_DOWN = 1073741905

VK_ESCAPE = 27
VK_BACKSPACE = 8
VK_ENTER = 13
VK_TAB = 9
VK_SPACE = 32

VK_CAPS_LOCK = 1073741881
VK_SCROLL_LOCK = 1073741895
VK_PAUSE = 1073741896

VK_AT = 64  # @
VK_PERCENT = 37  # %
VK_AMPERSAND = 38  # &
VK_TILDE = 126  # ~
VK_BAR = 124  # |
VK_UNDERSCORE = 95  # _
VK_CARET = 94  # ^
VK_DOLLAR = 36  # $
VK_HASH = 35  #
VK_BACKSLASH = 92  # \
VK_SHARP = VK_HASH
VK_UNDERBAR = VK_UNDERSCORE

VK_BANG = 33  # !
VK_QUESTION = 63  # ?
VK_ACCENT = 96  # `
VK_PERIOD = 46  # .
VK_COMMA = 44  # ,
VK_QUOTE = 96  # "
VK_APOSTROPHE = 39  # '

VK_HYPHEN = 45  # -
VK_COLON = 58  # :
VK_SEMICOLON = 59  # ;

VK_PARENTHESES_L = 40  # (
VK_PARENTHESES_R = 41  # )
VK_BRACE_L = 123  # {
VK_BRACE_R = 125  # }
VK_BRACKET = 91  # [
VK_BRACKET = 93  # ]
VK_BRACKET_ANGLE_L = 60  # <
VK_BRACKET_ANGLE_R = 62  # >

VK_EQUAL = 61  # =
VK_PLUS = 43  # +
VK_MINUS = VK_HYPHEN
VK_ASTERISK = 42  # *
VK_SLASH = 47  # /

VK_LESS = VK_BRACKET_ANGLE_L
VK_GREAT = VK_BRACKET_ANGLE_R
VK_MULTIPLY = VK_ASTERISK
VK_DIVIDE = VK_SLASH
VK_XOR = VK_CARET

VK_INSERT = 1073741897
VK_DELETE = 127
VK_HOME = 1073741898
VK_END = 1073741901
VK_PAGE_UP = 1073741899
VK_PAGE_DOWN = 1073741902

VK_A = 65
VK_B = 66
VK_C = 67
VK_D = 68
VK_E = 69
VK_F = 70
VK_G = 71
VK_H = 72
VK_I = 73
VK_J = 74
VK_K = 75
VK_L = 76
VK_M = 77
VK_N = 78
VK_O = 79
VK_P = 80
VK_Q = 81
VK_R = 82
VK_S = 83
VK_T = 84
VK_U = 85
VK_V = 86
VK_W = 87
VK_X = 88
VK_Y = 89
VK_Z = 90

VK_SA = 97
VK_SB = 98
VK_SC = 99
VK_SD = 100
VK_SE = 101
VK_SF = 102
VK_SG = 103
VK_SH = 104
VK_SI = 105
VK_SJ = 106
VK_SK = 107
VK_SL = 108
VK_SM = 109
VK_SN = 110
VK_SO = 111
VK_SP = 112
VK_SQ = 113
VK_SR = 114
VK_SS = 115
VK_ST = 116
VK_SU = 117
VK_SV = 118
VK_SW = 119
VK_SX = 120
VK_SY = 121
VK_SZ = 122

VK_F01 = 1073741882
VK_F02 = 1073741883
VK_F03 = 1073741884
VK_F04 = 1073741885
VK_F05 = 1073741886
VK_F06 = 1073741887
VK_F07 = 1073741888
VK_F08 = 1073741889
VK_F09 = 1073741890
VK_F10 = 1073741891
VK_F11 = 1073741892
VK_F12 = 1073741893


class INPUT_STATES:
    NONE = 0
    PRESSED = 1
    ING = 2
    RELEASED = 3


class MASKS:
    RECTANGLE = 0
    PRECISE = 1
    CIRCLE = 2



from typing import Optional, overload

from .preset import RsPreset
from .scene import RsScene


def room_register(info: RsScene, caption: Optional[str] = None):
    NewRoom: RsScene = info
    Name: str
    if caption:
        if caption == "":
            raise RuntimeError(f"Please enter name of the scene {id(info)}")
        NewRoom.name = caption
        Name = caption
    else:
        Name = info.name

    Number = len(RsPreset.room_order)
    if 0 < Number:
        LastRoom = RsPreset.room_last
        if LastRoom and NewRoom:
            NewRoom.before = LastRoom
            LastRoom.next = NewRoom
    else:
        RsPreset.room = NewRoom
    RsPreset.room_last = NewRoom

    RsPreset.room_order.append(NewRoom)
    RsPreset.room_all[Name] = NewRoom
    return NewRoom


@overload
def room_lookup(id: int) -> Optional[RsScene]: ...


@overload
def room_lookup(id: str) -> Optional[RsScene]: ...


def room_lookup(id):
    if isinstance(id, int):
        return RsPreset.room_order[id]
    elif isinstance(id, str):
        return RsPreset.room_all[id]


def _room_set(taget: RsScene):
    if RsPreset.room:
        RsPreset.room.onDestroy()
    RsPreset.room = taget
    RsPreset.room.onAwake()

    if RsPreset.debug_get():
        print("Go to " + str(RsPreset.room))


def room_goto(name: str):
    Temp = room_lookup(name)
    if not Temp:
        raise RuntimeError("The room " + name + " doesn't exist.")
    elif Temp is not RsPreset.room:
        _room_set(Temp)


def room_goto_next():
    Next = RsPreset.room.next
    if Next:
        _room_set(Next)
    else:
        raise RuntimeError("The next room doesn't exist.\n")



""" Sprite
    ---
    ```
    from realpy import RsSprite
    ```
"""
from typing import Optional

from pygame.surface import Surface

__all__ = ["RsSprite"]


class RsSprite(object):
    """`RsSprite(image, mask_type, x_offset, y_offset)`
        ---
        Advance asset of image.
    """

    __slots__ = [
        "image", "mask_type", "width", "height", "radius", "xoffset", "yoffset", "boundbox", "center_distance",
        "center_angle", "using_custom_offset"
    ]

    from realpy.core.image import RsImage

    def __init__(self, image: RsImage, mask_type: int = 0, xo: int = 0, yo: int = 0):
        from realpy.core.image import RsImage
        from realpy.core.arithmetic import point_distance, point_direction

        assert image
        self.image: RsImage = image
        self.mask_type: int = mask_type
        self.width: int = image.boundbox.width
        self.height: int = image.boundbox.height
        self.radius: int = max(self.width, self.height)
        self.xoffset: int = xo
        self.yoffset: int = yo
        self.using_custom_offset: bool = False

        HalfX, HalfY = int(self.width * 0.5), int(self.height * 0.5)

        if self.xoffset == HalfX and self.yoffset == HalfY:
            self.center_distance = 0
            self.center_angle = 0
        else:
            self.center_distance = point_distance(xo, yo, HalfX, HalfY)
            self.center_angle = point_direction(xo, yo, HalfX, HalfY)
            self.using_custom_offset = True

        bx, bex = -xo, self.width - xo
        by, bey = -yo, self.height - yo
        self.boundbox = [(bx, by), (bex, by), (bx, bey), (bex, bey)]

    def draw(self, where: Surface, index, x, y, scale: float = 1, orientation: float = 0,
             alpha: float = 1, *, xflip: bool = False, yflip: bool = False) -> Optional[Surface]:
        from pygame import transform
        from pygame.surface import Surface
        from realpy.core.arithmetic import lengthdir_x, lengthdir_y

        if scale <= 0 or alpha <= 0:
            return None
        elif self.image:
            if 0 < self.image.number:
                index = index % self.image.number
                Frame: Surface = self.image.raw_data[index]

                Sizes = (int(scale * Frame.get_width()), int(scale * Frame.get_height()))
                Trx: Surface
                Trx = transform.scale(Frame, Sizes)
                if xflip or yflip:
                    Trx = transform.flip(Trx, xflip, yflip)

                if orientation != 0:
                    Trx = transform.rotate(Trx, orientation)
                # Trx = transform.rotozoom(Frame, orientation, scale)

                Position = Trx.get_rect()
                if self.using_custom_offset:
                    Lx = -lengthdir_x(self.center_distance, self.center_angle + orientation) * scale
                    if xflip:
                        Lx *= -1
                    Ly = -lengthdir_y(self.center_distance, self.center_angle + orientation) * scale
                    if yflip:
                        Ly *= -1
                    Position.center = (x + Lx, y + Ly)
                else:
                    Position.center = (x, y)

                where.blit(Trx, Position)
                return Trx
        return None


"""
def sprite_json_loads():
    try:
        with open("Data\\sprite.json") as sprfile:
            parsed = json.load(sprfile)

            for content in parsed:
                sprite_load(content["path"], content["name"], content["xoffset"], content["yoffset"], content["number"])

    except FileNotFoundError:
        pass
"""



from typing import List, Optional

from pygame import mixer as PyAudio
from pygame.mixer import Channel as PyChannel, Sound as PySound


class RsSound(PySound):
    def __init__(self, info):
        super().__init__(info)
        self.fields: List[PyChannel] = []

    def play(self, loops: Optional[int] = 0, maxtime: Optional[int] = 0, fade_ms: Optional[int] = 0) -> Optional[
        PyChannel]:
        Place: Optional[PyChannel] = PyAudio.find_channel(False)
        if not Place:
            return None
        else:
            self.fields.append(Place)
            Place.play(self, loops, maxtime, fade_ms)
            return Place

    def stop(self) -> None:
        self.fields.clear()
        super().stop()

    def is_playing(self) -> bool:
        return 0 < self.get_num_channels()



from typing import List


class RsActor(object):
    """`RsActor(Layer)`
        ---
    """

    def __init__(self, layer):
        self.enabled: bool = True
        self.layer = layer
        self.department: List[List] = []
        self.onAwake = None
        self.onDestroy = None
        self.onUpdate = None
        self.onUpdateLater = None
        self.onDraw = None

    def __del__(self):
        for Group in self.department:
            Group.remove(self)

    def attach(self, ogroup: List):
        ogroup.append(self)
        self.department.append(ogroup)



def test_main():
    print("***** Test → Realpy Framework (+Prefab and Assets) *****")

    try:
        import asyncio
        from realpy import RsFramework, RsScene, RsLayer, RsPrefab
        from realpy import room_register, instance_create

        print(">>> Module of machine: ", RsFramework)
        print(">>> Class of scene: ", RsScene)
        print(">>> Class of layer: ", RsLayer)

        Sample = RsScene()
        SampleLayers = [RsLayer("First"), RsLayer("Second"), RsLayer("Third")]
        print("Sample scene: ", Sample)
        print("Sample layer group: ", SampleLayers)

        for Item in SampleLayers:
            Sample.add_layer_direct(Item)
        SampleSeek = Sample.layer_find("First")
        print("Found scene (maybe 'First'): ", SampleSeek)
        SampleSeek = Sample.layer_find("Fourth")
        print("Found scene (maybe 'None'): ", SampleSeek)

        Sample.add_layer_direct(RsLayer("Fifth"))
        print("The layers: ", Sample.layer_stack)

        Sample.onAwake()

        async def timeout():
            await asyncio.sleep(5)
            print("x of - Sample Instance 2:", TestInstance2.x)
            raise TimeoutError

        class oSampleObject1(RsPrefab):
            @staticmethod
            def onAwake(itself):
                print("Instance 1: ", id(itself))

            @staticmethod
            def onDestroy(itself):
                print("Instance 1 died! - ", id(itself))

            @staticmethod
            def onUpdate(itself, time: float):
                itself.x += 5 * time

        class oSampleObject2(RsPrefab):
            @staticmethod
            def onAwake(itself):
                print("Instance 2: ", id(itself))

            @staticmethod
            def onDestroy(itself):
                print("Instance 2 died! - ", id(itself))

            @staticmethod
            def onUpdate(itself, time: float):
                print("Instance 2 -> middle")

            @staticmethod
            def onUpdateLater(itself, time: float):
                print("Instance 2 -> later")

        print(">>> Class of prefab #1: ", oSampleObject1)
        print(">>> Class of prefab #2: ", oSampleObject2)

        RsFramework.rs_init("RealPy Engine", 200, 200)
        asyncio.run(timeout())

        TestRoom = room_register(RsScene("roomTest"))
        print(repr(TestRoom))

        Testbed = TestRoom.add_layer_direct(RsLayer("Instances"))
        print(repr(Testbed))

        TestInstance1 = instance_create(oSampleObject1, Testbed)
        TestInstance2 = instance_create(oSampleObject2, Testbed)
        TestInstance3 = instance_create(oSampleObject2, Testbed)
        print("Sample Instance 1:", TestInstance1)
        print("Sample Instance 2:", TestInstance2)
        print("Sample Instance 3:", TestInstance3)
        print("The instances: ", Testbed.storage)

        RsFramework.rs_startup()
    except AttributeError as e:
        print("Attribute Error: ", e)
        return False
    except SystemError as e:
        print("System Error: ", e)
        return False
    except ValueError as e:
        print("Value Error: ", e)
        return False
    except RuntimeError as e:
        print("Runtime Error: ", e)
        return False
    except TimeoutError:
        return True
    except Exception as e:
        print("Error: ", e)
        return False
    finally:
        print("***** Realpy Test Finished *****")
        return True


if __name__ == "__main__":
    assert test_main()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="realpy-engine-iconer",  # Replace with your own username
    version="1.0.0",
    author="Iconerworks",
    author_email="yoyofa2@hotmail.com",
    description="2D game engine with pygame",
    license="LICENSE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RealPy-Engine/realpy-core",
    project_urls={
        "Bug Tracker": "https://github.com/RealPy-Engine/realpy-core/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
)



from .arithmetic import *
from .behavior import *
from .collisionfield import *
from .constants import *
from .framework import *
from .function_input import *
from .function_instance import *
from .function_scene import *
from .image import *
from .layer import *
from .preset import *
from .scene import *
from .sound import *
from .sprite import *
